/**
 ** nddExecTaskCodels.c
 **
 ** Codels appeles par la tache nddExecTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#include <sys/time.h>

#ifndef UNIX
#include <vxWorks.h>
#else
#include <portLib.h>
#endif
#include <stdio.h>
#include "server/nddHeader.h"

#include <strings.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include <h2mathLib.h>
#include <posterLib.h>

/* From the sick module */
#include "sickConst.h"
#include "sickStruct.h"
#include "sick_convert.h"

/* From the aspect module */
#include "aspectStruct.h"

/* From mapkit */
#include <aspectLib.h>

/* From the pom module */
#include "pomStruct.h"
#include "libeuler/pomEuler.h"

#include "nddStruct.h"
#include "nddConst.h"
#include "lib/nddLib.h"
#include "lib/nddLibUtils.h"

#define NDD_GOTOING 1
#define NDD_TRACKING 2
#define NDD_FINISHING 3
#define NDD_BLOCKED 4
#define NDD_FINAL_ROTATING 5
#define NDD_NOT_MOVING 0

static SICK_SCANCART_POSTER_STR nddSickPoster;
static SICK_SCANPOLAR_POSTER_STR nddSickPolarPoster;
static ASPECT_POLARASPECT_POSTER_STR nddPolarAspectPoster;
static ASPECT_CARTESIANASPECT_POSTER_STR nddCartAspectPoster;

static PILO_TRAJECTORY nddPiloTrajectory;

static POSTER_ID obstPosterId;  
static POSTER_ID currentPosPosterId;
static POSTER_ID trackPosterId;

static int testBlockedDuration = 0;

static NDD_DIAGRAM diagrams;

/* The width of the security zone arround the robot depends on the speed
   of the robot. By default the enlargement of this zone is computed for 
   the maximal speed vmax (more or les the distance reach by the robot in 
   NB_MAX_ITER_ENLARGMENT ndd steps). If vmax is big then enlargement could 
   be much more too large for small speeds. Thus the NB_ENLARGMENTS allows 
   to computes as many enlargments discretizing the speed values between 
   ]0 : vmax]. However some experimentations show motion saccades that seem 
   related to the switch between enlargement values. So for now we use 
   only one enlargement zone. */
#define NB_ENLARGMENTS 1
#define NB_MAX_ITER_ENLARGMENT 15
static float * enlargement[NB_ENLARGMENTS];

static int nddState = NDD_NOT_MOVING;
static int nddWasBlocked = FALSE;
static int nddStopRequired = FALSE;
static int nddStopReport = OK;

static GENPOS_CART_CONFIG * currentTraj = NULL;

/*local functions to connect to nddLib (genom-dependant) */
static BOOL getLocalGoal(NDD_POINT *);
static STATUS readObstPoster(POSTER_ID srcPosterId, NDD_OBST_TYPE stype, int *bilan);
static STATUS updateObstacles(int *bilan);

static void setSpeedrefPosterToZero();
static int setSpeedrefPosterToZeroSmoothly();
/* static void finishedAllMotions(); */
static double getDistToGoal();
static STATUS updateCurrentPos(int *report);
static void setRobotOrientation();
static int updateTrajGoalPos();
static STATUS nddGotoLocal(int * bilan, double vmax, double wmax);
static double getDistToPoint(double x, double y);

static BOOL testFreshData(unsigned int date, int *bilan);

static void nddReport(int *bilan);
static void resestTestBlocked(void);
static BOOL testBlocked(void);
static void filterSpeed(NDD_SPEED_REF *wantedSpeed, GENPOS_CART_SPEED *oldSpeed);

/*------------------------------------------------------------------------
 *
 * nddExecTaskInit  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS nddExecTaskInit(int *bilan)
{
  int i;

  SDI_F->inputPosters.obstType = NDD_DEFAULT_OBST_TYPE;
  strncpy(SDI_F->inputPosters.obstPosterName,
	  NDD_DEFAULT_OBST_POSTER,
	  strlen(NDD_DEFAULT_OBST_POSTER));

  strncpy(SDI_F->trajPosterName.name,
	  NDD_DEFAULT_TRAJ_POSTER,
	  strlen(NDD_DEFAULT_TRAJ_POSTER));

  SDI_F->maxSpeed.linear = NDD_MAX_LIN_SPEED;
  SDI_F->maxSpeed.angular = NDD_MAX_ANG_SPEED;

  SDI_F->params.distToGoal = NDD_DEFAULT_DIST_TO_GOAL;
  SDI_F->params.securityDistance = NDD_DEFAULT_SECURITY_DISTANCE;
  SDI_F->params.smax = NDD_DEFAULT_SMAX;
  SDI_F->params.lmaxRobot = NDD_DEFAULT_LMAX_ROBOT;
  SDI_F->params.dMaxLocalMap = NDD_DEFAULT_DMAXLOCALMAP;
  SDI_F->params.epsilonTheta = NDD_DEFAULT_EPSILON_THETA;
  SDI_F->params.useTheta = NDD_USE_THETA;
  SDI_F->params.distSwitchTraj = NDD_DEFAULT_DISTSWITCHTRAJ;
  SDI_F->params.decelMax = NDD_DEFAULT_MAXDECEL;
  SDI_F->params.enlargementFactor =  NDD_DEFAULT_ENLARGEMENT_FACTOR;
  SDI_F->params.linFactor = NDD_DEFAULT_LIN_FACTOR;

  SDI_F->global_goal.x = 0;
  SDI_F->global_goal.y = 0;
  SDI_F->global_goal.theta = 0;

  SDI_F->local_goal.x = 0;
  SDI_F->local_goal.y = 0;
  SDI_F->local_goal.theta = 0;

  SDI_F->motionChoice.motionDirection =  0;
  SDI_F->motionChoice.valley =  0;
  SDI_F->motionChoice.speedPercent = 0.0;

  SDI_F->speedRef.numRef = 0;
  SDI_F->speedRef.updatePeriod = NDD_TICS_PERIOD;
  SDI_F->speedRef.v = 0.0;
  SDI_F->speedRef.vt = 0.0;
  SDI_F->speedRef.w = 0.0;
  SDI_F->speedRef.vmax = NDD_MAX_LIN_SPEED;
  SDI_F->speedRef.wmax = NDD_MAX_ANG_SPEED;
  SDI_F->speedRef.linAccelMax = NDD_MAX_LIN_ACCEL;
  SDI_F->speedRef.angAccelMax = NDD_MAX_ANG_ACCEL;
#if defined (XR4000) /* XXXXX */
   SDI_F->speedRef.globalFlag = XR4000_JOINT;
#endif

  SDI_F->nextToVisit = -1;

  diagrams.nsectors = NDD_NSECTORS;
  diagrams.lmax_robot = SDI_F->params.lmaxRobot;
  diagrams.dmax_sensor = SICK_SEG_RHO_MAX;
  diagrams.currentStrat = NDDLIB_STRAT_NO;
  
  if (enlargement[0]) {
    for (i=0; i<NB_ENLARGMENTS; i++)
      free(enlargement[i]);
    enlargement[0] = NULL;
  }

  for (i=0; i<NB_ENLARGMENTS; i++)
    enlargement[i] = nddComputeEnlargement(NDD_NSECTORS,
					   SDI_F->params.lmaxRobot,
					   SDI_F->params.securityDistance,
					   SDI_F->maxSpeed.linear,
					   SDI_F->params.enlargementFactor*EXEC_TASK_PERIOD(NDD_EXECTASK_NUM));

  nddState = NDD_NOT_MOVING;
/*   fprintf(stdout, "ndd State changed to NDD_NOT_MOVING\n"); */

  return OK;
}

/*------------------------------------------------------------------------
 *
 * nddExecTaskEnd  --  Codel de terminaison
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS nddExecTaskEnd(void)
{
  setSpeedrefPosterToZero();
  return OK;
}

/*------------------------------------------------------------------------
 *
 * nddExecTaskPerm  --  codel de l'activite' permanente
 *
 * Description: 
 * 
 * Bilans:      OK 
 *              S_ndd_E_EXEC_TASK_GENERAL_FAILURE
 * 
 * Retourne:    OK ou ERROR
 */

STATUS nddExecTaskPerm(int *bilan)
{
  int last =1;
  double distStop;
  NDD_POINT local_goal;

  /* update pos and local env */
  if (updateObstacles(bilan) == ERROR) {
    nddStopRequired = TRUE;
    nddState = NDD_FINISHING;
  }

  if (nddState == NDD_NOT_MOVING) {
    resestTestBlocked();
    return OK;
  }
  
  if (nddState == NDD_FINAL_ROTATING) {
      if (SDI_F->params.useTheta == NDD_USE_THETA) { 
        if (fabs(angleLimit(SDI_F->current_position.mainToOrigin.euler.yaw 
			    - SDI_F->global_goal.theta)) 
	    > SDI_F->params.epsilonTheta) {
	  setRobotOrientation();  
	}
        else {
	  nddState = NDD_NOT_MOVING;
	  setSpeedrefPosterToZero();
	  fprintf(stdout, "ndd State changed from ORIENTING ROBOT to NDD_NOT_MOVING\n"); 
	}
      } else {
	nddState = NDD_NOT_MOVING;
	fprintf(stdout, "ndd State changed from FINAL_ROTATING to NDD_NOT_MOVING\n"); 
      }
    return OK;
  }

  /* end of motion */
  if ( (nddState == NDD_BLOCKED) || (nddState == NDD_FINISHING)) {
    getLocalGoal(&local_goal);
    if (setSpeedrefPosterToZeroSmoothly()) {

      /* final rotating */
      if (!nddStopRequired && nddState == NDD_FINISHING && SDI_F->params.useTheta == NDD_USE_THETA) {
	  nddState = NDD_FINAL_ROTATING;
	  fprintf(stdout, "ndd State changed to NDD_FINAL_ROTATING\n");
      } else {
	nddState = NDD_NOT_MOVING;
	fprintf(stdout, "ndd State changed to NDD_NOT_MOVING\n");
      }
    }
    return OK;
  }

  /* 
   * Test if blocked
   */
  if (nddState !=  NDD_FINAL_ROTATING)
    testBlocked();

  
  /* 
   * Update goal
   */
  if (nddState == NDD_TRACKING) 
    last = updateTrajGoalPos();
  if (SDI_F->params.useTheta == NDD_USE_THETA)
    angleLimit(SDI_F->global_goal.theta);


  /* 
   * Dist required to stop
   */
  distStop = MAX(SDI_F->speedRef.v * SDI_F->speedRef.v 
		 / SDI_F->params.decelMax / 2.0,
		 SDI_F->params.distToGoal);

#if VERBOSE > 0
  fprintf(stdout, "ndd: next absolute goal x: %5.2f  y: %5.2f   distStop=%5.2f\n", 
	  SDI_F->global_goal.x, SDI_F->global_goal.y, distStop);
#endif


  /* last point */
  if (last && (getDistToGoal() < distStop)) {
    fprintf(stdout, "ndd: FINAL GOAL REACHED !!\n");
   
    /* allows to restart with same traj */
    if (currentTraj) free(currentTraj);
    currentTraj = NULL;
    nddState = NDD_FINISHING;
    printf ("end traj resestTestBlocked\n");
    resestTestBlocked();
/*     fprintf(stdout, "ndd State changed to NDD_FINISHING\n"); */
  } else 
    nddGotoLocal(bilan,SDI_F->maxSpeed.linear,SDI_F->maxSpeed.angular);
  
  return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_ndd_E_UNKNOW_OBST
 */

/* nddInitExec  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
nddInitExec(NDD_INPUT_POSTERS *inputPosters, int *report)
{
  if ( SDI_F->inputPosters.obstType != NDD_INPUT_DEBUG)
    updateObstacles(report);

  updateCurrentPos(report);

  return ETHER;
}

/*------------------------------------------------------------------------
 * SetParams
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_ndd_E_UNKNOW_POSTER
 */

/* nddSetParamsExec  -  codel EXEC de SetParams
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddSetParamsExec (NDD_PARAMS *(in_params), int *bilan)
{
  int i;

  diagrams.nsectors = NDD_NSECTORS;
  diagrams.lmax_robot = SDI_F->params.lmaxRobot;
  diagrams.dmax_sensor = SICK_SEG_RHO_MAX;
  
  if (enlargement[0]) {
    for (i=0; i<NB_ENLARGMENTS; i++)
      free(enlargement[i]);
    enlargement[0] = NULL;
  }
  for (i=0; i<NB_ENLARGMENTS; i++)
    enlargement[i] = nddComputeEnlargement(NDD_NSECTORS,
					   SDI_F->params.lmaxRobot,
					   SDI_F->params.securityDistance,
					   SDI_F->maxSpeed.linear,
					   SDI_F->params.enlargementFactor*EXEC_TASK_PERIOD(NDD_EXECTASK_NUM));

  if (SDI_F->inputPosters.obstType == NDD_INPUT_CART_POINTS ||
      SDI_F->inputPosters.obstType == NDD_INPUT_POLAR_POINTS)  {      
    if (posterFind(SDI_F->inputPosters.obstPosterName, &obstPosterId) == ERROR) 
      *bilan = errnoGet();
  } else if (SDI_F->inputPosters.obstType == NDD_INPUT_DEBUG) 
    *bilan = OK;

  nddParamsParamsPosterWrite(NDD_PARAMS_POSTER_ID, in_params);

    return ETHER;
}

/*------------------------------------------------------------------------
 * SetSpeed
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_ndd_E_UNKNOW_POSTER
 */

/* nddSetSpeedStart  -  codel START de SetSpeed
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddSetSpeedExec (NDD_SPEED *(in_speed), int *bilan)
{
  int i;

  if (enlargement[0]) {
    for (i=0; i<NB_ENLARGMENTS; i++)
      free(enlargement[i]);
    enlargement[0] = NULL;
  }

  for (i=0; i<NB_ENLARGMENTS; i++)
    enlargement[i] = nddComputeEnlargement(NDD_NSECTORS,
					   SDI_F->params.lmaxRobot,
					   SDI_F->params.securityDistance,
					   in_speed->linear,
 					   SDI_F->params.enlargementFactor*EXEC_TASK_PERIOD(NDD_EXECTASK_NUM));

  SDI_F->speedRef.vmax = in_speed->linear;
  SDI_F->speedRef.wmax = in_speed->angular;
  SDI_F->speedRef.linAccelMax = in_speed->accelLinear;
  SDI_F->speedRef.angAccelMax = in_speed->accelAngular;

  nddMaxSpeedMaxSpeedPosterWrite(NDD_MAXSPEED_POSTER_ID, in_speed);
  return ETHER;
}

/*------------------------------------------------------------------------
 * ExecTraj
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_ndd_E_UNKNOW_POSTER
 */

/* nddExecTrajStart  -  codel START de ExecTraj
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddExecTrajStart (NDD_STRING *trajPosterName, int *bilan)
{
  if (posterFind(trajPosterName->name, &trackPosterId) == ERROR) {
    *bilan = errnoGet();
  }   
  nddStopRequired = FALSE;

  fprintf(stdout, "ndd:: TRACK !!!!\n");
  nddState = NDD_TRACKING;
  fprintf(stdout, "ndd State changed to NDD_TRACKING\n");
  return EXEC;
}

/* nddExecTrajExec  -  codel EXEC de ExecTraj
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddExecTrajExec (NDD_STRING *trajPosterName, 
				int *bilan)
{
  if (nddState == NDD_NOT_MOVING) {
    nddReport(bilan);
    return ETHER;
  }

  return EXEC;
}

/* nddExecTrajInter  -  codel INTER de ExecTraj
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddExecTrajInter (NDD_STRING *trajPosterName, 
				 int *bilan)
{
  nddStopRequired = TRUE;
  printf ("NDD: TRAJ INTER\n");

  if (nddState == NDD_NOT_MOVING){
    nddReport(bilan);
    return ETHER;
  }

  nddState = NDD_FINISHING;
  fprintf(stdout, "ndd State changed to NDD_FINISHING\n"); 

  return INTER;
}


/*------------------------------------------------------------------------
 * GoTo
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_ndd_E_BLOCKED
 */

/* nddGoToStart  -  codel START de GoTo
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddGoToStart (GENPOS_CART_CONFIG *goal, int *bilan)
{
  nddStopRequired = FALSE;
  nddState = NDD_GOTOING;
  goal->theta = fmod(goal->theta, M_PI);
/*   fprintf(stdout, "ndd State changed to NDD_GOTOING\n"); */
  fprintf(stdout,"ndd: GOTO!!!!\n");
  return EXEC;
}

/* nddGoToExec  -  codel EXEC de GoTo
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddGoToExec (GENPOS_CART_CONFIG *goal, int *bilan)
{
  if (nddState == NDD_NOT_MOVING) {
    nddReport(bilan);
    return ETHER;
  }
  return EXEC;
}

/* nddGoToInter  -  codel INTER de GoTo
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddGoToInter (GENPOS_CART_CONFIG *goal, int *bilan)
{
  nddStopRequired = TRUE;
  nddState = NDD_FINISHING;
  nddReport(bilan);
  printf ("NDD: GOTO INTER\n");
/*   fprintf(stdout, "ndd State changed to NDD_FINISHING\n"); */
  return ETHER;
}



/*------------------------------------------------------------------------
 * GoToRelative
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_ndd_E_BLOCKED
 *              S_ndd_E_CANNOT_READ_SOURCE
 */

/* nddGoToRelativeStart  -  codel START de GoToRelative
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddGoToRelativeStart (GENPOS_CART_CONFIG *local_goal, int *bilan)
{
  double xR, yR, thR;
  nddStopRequired = FALSE;

  xR = SDI_F->current_position.mainToOrigin.euler.x;
  yR = SDI_F->current_position.mainToOrigin.euler.y;
  thR = SDI_F->current_position.mainToOrigin.euler.yaw;

  /* convert local to global */
  SDI_F->global_goal.x =  xR 
			  + local_goal->x * cos(thR) 
			  - local_goal->y * sin(thR);
  SDI_F->global_goal.y =  yR 
			  + local_goal->x * sin(thR) 
			  + local_goal->y * cos(thR);
  SDI_F->global_goal.theta = fmod(thR + local_goal->theta, M_PI);

  nddState = NDD_GOTOING;
/*   fprintf(stdout, "ndd State changed to NDD_GOTOING\n"); */
  fprintf(stdout,"ndd: GOTO Relative!!!!\n");
  return EXEC;
}

/* nddGoToRelativeExec  -  codel EXEC de GoToRelative
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddGoToRelativeExec (GENPOS_CART_CONFIG *local_goal, int *bilan)
{
  if (nddState == NDD_NOT_MOVING) {
    nddReport(bilan);
    return ETHER;
  }
  return EXEC;
}

/* nddGoToRelativeInter  -  codel INTER de GoToRelative
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT nddGoToRelativeInter (GENPOS_CART_CONFIG *local_goal, int *bilan)
{
  nddStopRequired = TRUE;
  printf ("NDD: GOTO REL INTER\n");
  nddState = NDD_FINISHING;
  nddReport(bilan);

/*   fprintf(stdout, "ndd State changed to NDD_FINISHING\n"); */
  return ETHER;
}

/******************************************************/
/*         internal functions                         */
/******************************************************/


static void setSpeedrefPosterToZero()
{
  printf ("setSpeedrefPosterToZero\n");
  SDI_F->speedRef.numRef++;
  SDI_F->speedRef.w = 0.0;
  SDI_F->speedRef.v = 0.0;
  SDI_F->speedRef.vt = 0.0;
}

static int setSpeedrefPosterToZeroSmoothly() {
  double newV, newW, dV;
  int bilan;
  newV = SDI_F->speedRef.v;
  dV = SDI_F->params.decelMax*EXEC_TASK_PERIOD(NDD_EXECTASK_NUM);
  if (dV < 0.02) {
    printf ("dV %g too small !!!! set to 0.02\n", dV);
    dV = 0.02;
  }
  newV -= dV;
  newW = newV/SDI_F->speedRef.v * SDI_F->speedRef.w;

  fprintf(stdout,"NDD: speeding down of %g to stop smoothly (period = %g, decelMax %g) newV = %g\n", 
	  dV, EXEC_TASK_PERIOD(NDD_EXECTASK_NUM), SDI_F->params.decelMax, newV);

  if (newV <= dV+0.001) {
    setSpeedrefPosterToZero();
    return 1;
  }

  
  nddGotoLocal(&bilan, newV, newW);

/*   SDI_F->speedRef.numRef++; */
/*   SDI_F->speedRef.w = newV/SDI_F->speedRef.v * SDI_F->speedRef.w; */
/*   SDI_F->speedRef.v = newV; */
/*   SDI_F->speedRef.vt = 0.0; */

  SDI_F->motionChoice.motionDirection =  0;
  SDI_F->motionChoice.valley =  0;
  if (SDI_F->maxSpeed.linear != 0)
    SDI_F->motionChoice.speedPercent = SDI_F->speedRef.v/SDI_F->maxSpeed.linear;
  else
    SDI_F->motionChoice.speedPercent = SDI_F->speedRef.v;

  return 0;
}

/* ----------------------------------------------------------------------
 *
 * readObstPoster - get obstacles
 *
 * XXXX MUST BE PUT ON THE FRAME OF THE REAL CURRENT ROBOT POSITION XXX
 *
 * (see updateCurrentPos())
 *
 */

static STATUS readObstPoster(POSTER_ID srcPosterId, NDD_OBST_TYPE stype, 
			       int *bilan) 
{  
  int i;
  POM_EULER_V t3dAspectToGlobal, t3dAspectToCurrent; 
  int reversedSegs[SICK_MAX_ACQ_SEGS];
  double dtheta;

  /* read obstacles according to data type */
  switch (stype) {

  /* --
     -- cartesian points 
     -- */
  case NDD_INPUT_CART_POINTS:

    if (nddSICK_SCANCART_POSTER_STRPosterRead(srcPosterId, &nddSickPoster)
	== ERROR) {
	fprintf (stderr, "ndd: could not read sickPoster\n");
	*bilan = S_ndd_E_CANNOT_READ_SOURCE;
	return ERROR;
      }

    if (!testFreshData(nddSickPoster.Header.sickPomPos.date, bilan)) {
      fprintf(stderr, "ndd: sickPoster not fresh\n");
      return ERROR;
    }

    /* Convert to robot frame if required */
    sickConvertPtsFrame(nddSickPoster.Header.frame,
			SICK_ROBOT_FRAME,
			&nddSickPoster.Header.sickPomPos,
			nddSickPoster.Header.np,
			nddSickPoster.Points,
			nddSickPoster.Points);
    nddSickPoster.Header.frame = SICK_ROBOT_FRAME;

    if (updateCurrentPos(bilan) == ERROR)
      return ERROR;

    return OK;
    
  /* --
     -- polar points 
     -- */
    /* After this transformation, the data will be NDD_INPUT_CART_POINTS */
    /* However, we cannot change the type here because it will 
       then not be anymore correct for the next iteration */
  case NDD_INPUT_POLAR_POINTS:

    if (nddSICK_SCANPOLAR_POSTER_STRPosterRead(srcPosterId, &nddSickPolarPoster)
	== ERROR) {
	fprintf (stderr, "ndd: could not read sickPoster\n");
	*bilan = S_ndd_E_CANNOT_READ_SOURCE;
	return ERROR;
      }

    if (!testFreshData(nddSickPolarPoster.Header.sickPomPos.date, bilan)) {
      fprintf(stderr, "ndd: sickPoster not fresh\n");
      return ERROR;
    }


    /* NDD_INPUT_POLAR_POINTS -> NDD_INPUT_CART_POINTS */
    sickPolarToCart2(nddSickPolarPoster.Header.np,
		     &nddSickPolarPoster.Header.sickPomPos,
		     SICK_ROBOT_FRAME,
		     nddSickPolarPoster.Rho,
		     &nddSickPolarPoster.Header.theta_min,
		     &nddSickPolarPoster.Header.dtheta,
		     nddSickPoster.Points);
    
    nddSickPoster.Header.frame = SICK_ROBOT_FRAME;
    nddSickPoster.Header.sickPomPos = nddSickPolarPoster.Header.sickPomPos;
    nddSickPoster.Header.np = nddSickPolarPoster.Header.np;

    if (updateCurrentPos(bilan) == ERROR)
      return ERROR;

    return OK;
    

  /* --
     -- polar segments 
     -- */
  case NDD_INPUT_POLAR_SEGS:

    if (nddASPECT_POLARASPECT_POSTER_STRPosterRead(srcPosterId, 
						   &nddPolarAspectPoster) 
	== ERROR) {
      fprintf (stderr, "ndd: could not read aspectPoster\n");
      *bilan = S_ndd_E_CANNOT_READ_SOURCE;
      return ERROR;
    }
    if (!testFreshData(nddPolarAspectPoster.Header.sickPomPos.date, bilan)) {
      fprintf(stderr, "ndd: sickPoster not fresh\n");
      return ERROR;
    }

    /* XXXXX BECARFUL OF THE FRAME !!!!!! */

    /* new robot pos */
    if (updateCurrentPos(bilan) == ERROR)
      return ERROR;

    dtheta = angleLimit(SDI_F->current_position.mainToOrigin.euler.yaw -
			nddPolarAspectPoster.Header.sickPomPos.mainToOrigin.euler.yaw);

    nddPolarAspectPoster.Header.sickPomPos.date = SDI_F->current_position.pomTickDate;
    nddPolarAspectPoster.Header.sickPomPos.VLocal = SDI_F->current_position.VLocal;
    nddPolarAspectPoster.Header.sickPomPos.mainToOrigin = SDI_F->current_position.mainToOrigin;
    nddPolarAspectPoster.Header.sickPomPos.mainToBase = SDI_F->current_position.mainToBase;


    if (fabs(dtheta) > 0.02) {
      printf ("!!! NDD DTHETA = %5.2f deg (pom=%lf data=%lf) \n", 
	      RAD_TO_DEG(dtheta),
	      SDI_F->current_position.mainToOrigin.euler.yaw,
	      nddPolarAspectPoster.Header.sickPomPos.mainToOrigin.euler.yaw);
      for (i=0; i<nddPolarAspectPoster.Header.nbSegs; i++) {
	nddPolarAspectPoster.PolarSegs.segs[i].t1 -= dtheta;
	nddPolarAspectPoster.PolarSegs.segs[i].t2 -= dtheta;
      }
    }

    return OK;

  /* --
     -- cart segments 
     -- */
  case NDD_INPUT_CART_SEGS:

    /* read aspect in cart frame */
    if (nddASPECT_CARTESIANASPECT_POSTER_STRPosterRead(srcPosterId, 
						       &nddCartAspectPoster) 
	== ERROR) {
      fprintf (stderr, "ndd: could not read aspectCartPoster\n");
      *bilan = S_ndd_E_CANNOT_READ_SOURCE;
      return ERROR;
    }
    if (!testFreshData(nddCartAspectPoster.Header.sickPomPos.date, bilan)) {
      fprintf(stderr, "ndd: aspectCartPoster not fresh\n");
      return ERROR;
    }

    /* -- convert to current robot pos */

    /* transf to global (if required) */
    sickGetT3DConvert(nddCartAspectPoster.Header.frame,
		      SICK_GLOBAL_FRAME,
		      &nddCartAspectPoster.Header.sickPomPos,
		      &t3dAspectToGlobal);

    /* new robot pos */
    if (updateCurrentPos(bilan) == ERROR)
      return ERROR;

    nddCartAspectPoster.Header.sickPomPos.date = SDI_F->current_position.pomTickDate;
    nddCartAspectPoster.Header.sickPomPos.VLocal = SDI_F->current_position.VLocal;
    nddCartAspectPoster.Header.sickPomPos.mainToOrigin = SDI_F->current_position.mainToOrigin;
    nddCartAspectPoster.Header.sickPomPos.mainToBase = SDI_F->current_position.mainToBase;

    /* comp transf aspect robot pos -> curretn robot pos */
    pomEulerVCompIn(&t3dAspectToCurrent, 
	      &t3dAspectToGlobal,
	      &SDI_F->current_position.mainToOrigin);
    
    /* convert segs frame aspect -> current */
    sickConvertSegs(&t3dAspectToCurrent,
		    nddCartAspectPoster.Header.nbSegs, 
		    nddCartAspectPoster.Segs.segs,
		    nddCartAspectPoster.Segs.segs);
    
    nddCartAspectPoster.Header.frame = SICK_ROBOT_FRAME;

    /* convert to polar */
    nddPolarAspectPoster.Header = nddCartAspectPoster.Header;
    aspectCartToPolarFilter((ASPECT_SEG *)nddCartAspectPoster.Segs.segs,
			    nddPolarAspectPoster.PolarSegs.segs,
			    5.0,
			    FALSE,
			    0, 0, reversedSegs);

    return OK;

  /* --
     -- debug 
     -- */
  case NDD_INPUT_DEBUG :
      nddSickPoster.Header.np = SICK_N_PT_MAX;
      nddSickPoster.Header.frame = SICK_ROBOT_FRAME;
      nddSickPoster.Header.send_time = 0;
      nddSickPoster.Header.send_time = 10;
      nddSickPoster.Header.theta_min = -M_PI/2;
      nddSickPoster.Header.dtheta = M_PI/90;
      nddSickPoster.Header.sickPomPos.mainToOrigin = pomEulerVIdentity;
      nddSickPoster.Header.sickPomPos.mainToOrigin.euler.yaw = 0;
      nddSickPoster.Header.sickPomPos.mainToOrigin.euler.x = 0;
      nddSickPoster.Header.sickPomPos.mainToOrigin.euler.y = 0;

      for (i = 0; i<=SICK_N_PT_MAX/2; i++) {
	  nddSickPoster.Points[i].x = 10.0;
	  nddSickPoster.Points[i].y = -26.0/(double)SICK_N_PT_MAX*(double)i + 10.0;
	}

      nddSickPoster.Points[0].x = 10.0;
      nddSickPoster.Points[0].y = 0.0;
	
      for (i = SICK_N_PT_MAX/2; i<SICK_N_PT_MAX; i++) { 
	nddSickPoster.Points[i].x = 10.0; 
	nddSickPoster.Points[i].y = -10.0/(double)SICK_N_PT_MAX*(double)i ; 
      }

      if (updateCurrentPos(bilan) == ERROR)
	return ERROR;

      break;

  default:
    *bilan = S_ndd_E_CANNOT_READ_SOURCE;
    return ERROR;
  } /* switch */

  return OK;
}

static BOOL testFreshData(unsigned int date, int *bilan)
{
  static unsigned int prevDate=-1;
  static int count=0;

  if ( (date == prevDate) && (nddState == NDD_GOTOING)) {
    printf("***** ndd : sick not updated %d times (date = %d)\n", count, date);
    count++;
  }
  else
    count = 0;
  if (count > 20) {
    *bilan = S_ndd_E_CANNOT_READ_OBSTACLES; 
    count = 0;
    return FALSE;
  }
  prevDate = date;
  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 * updateCurrentPos - get current robot pos
 *
 * If fact returns the pos linked to the acquired data in order to 
 * compute a goal position that is coherent with these obstacles.
 * XXXX However introduce a delay of at least 400ms. 
 * Should be better to use the real current pos after correction of 
 * the obstacle positions according to the robot motion.
 *
 */

static STATUS updateCurrentPos(int *report) 
{
  POM_POS current_position;
  POM_ME_POS current_me_position;
  static BOOL posPosterTypePOM_POS = TRUE;

  /* there is a position poster to read */
  if (SDI_F->inputPosters.currentPosPosterName[0] != '\0') {

    /* try to find it */
    if (!currentPosPosterId) {
      /* POM_POS type */
      if (nddPOM_POSPosterFind(SDI_F->inputPosters.currentPosPosterName, 
			       &currentPosPosterId) == ERROR) {
	if (!currentPosPosterId) {
	  fprintf(stderr, "ndd: cannot find POM_POS poster\n");
	  *report = S_ndd_E_CANNOT_READ_POSITION;
	  return ERROR;
	}
	/* POM_ME_POS type */
	if (nddPOM_ME_POSPosterFind(SDI_F->inputPosters.currentPosPosterName, 
				    &currentPosPosterId) == ERROR) {
	  fprintf(stderr, "ndd: cannot find POM_ME_POS poster\n");
	  currentPosPosterId = NULL;
	  *report = S_ndd_E_CANNOT_READ_POSITION;
	  return ERROR;
	}
	posPosterTypePOM_POS = FALSE;
      }
    }
  }

  /* read the pos poster */
  if (currentPosPosterId) {
    /* POM_POS type */
    if (posPosterTypePOM_POS) {
      if (nddPOM_POSPosterRead(currentPosPosterId, &(current_position)) == ERROR) {
	fprintf(stderr, "ndd: unable to read pomPos poster\n");
	currentPosPosterId = NULL;
	*report = S_ndd_E_CANNOT_READ_POSITION;
	return ERROR;
      }
    }
    /* POM_ME_POS type */
    else {
      if (nddPOM_ME_POSPosterRead(currentPosPosterId, &(current_me_position)) == ERROR) {
	fprintf(stderr, "ndd: unable to read pomMePos poster\n");
	currentPosPosterId = NULL;
	*report = S_ndd_E_CANNOT_READ_POSITION;
	return ERROR;
      }
      current_position.date = current_me_position.date1;
      current_position.pomTickDate = current_me_position.date1;
      current_position.mainToOrigin = current_me_position.main;
      current_position.mainToBase = current_me_position.main;
      current_position.VLocal = current_me_position.VLocal;
    }
  }

  /* record position */
  switch(SDI_F->inputPosters.obstType) {
  case NDD_INPUT_POLAR_POINTS:
  case NDD_INPUT_CART_POINTS:
  case NDD_INPUT_DEBUG:
    SDI_F->current_position.date = nddSickPolarPoster.Header.sickPomPos.date;
    SDI_F->current_position.pomTickDate = nddSickPolarPoster.Header.sickPomPos.date;
    SDI_F->current_position.VLocal = nddSickPolarPoster.Header.sickPomPos.VLocal;
    SDI_F->current_position.mainToOrigin = nddSickPolarPoster.Header.sickPomPos.mainToOrigin;
    SDI_F->current_position.mainToBase = nddSickPolarPoster.Header.sickPomPos.mainToBase;
    break;
  case NDD_INPUT_POLAR_SEGS:
  case NDD_INPUT_CART_SEGS:
    if (currentPosPosterId) {
      SDI_F->current_position = current_position;
    }
    else {
      SDI_F->current_position.date = nddPolarAspectPoster.Header.sickPomPos.date;
      SDI_F->current_position.pomTickDate = nddPolarAspectPoster.Header.sickPomPos.date;
      SDI_F->current_position.VLocal = nddPolarAspectPoster.Header.sickPomPos.VLocal;
      SDI_F->current_position.mainToOrigin = nddPolarAspectPoster.Header.sickPomPos.mainToOrigin;
      SDI_F->current_position.mainToBase = nddPolarAspectPoster.Header.sickPomPos.mainToBase;
    }
    break;


  default:
    fprintf(stderr, "ndd: unknown inputPosters type\n");
    *report = S_ndd_E_UNKNOW_OBST;
    return ERROR;
  }

#if 0
  printf ("----- dist aspect-real pos = %5.2f m %5.2f deg\n", 
	  sqrt(SQR(current_position.mainToOrigin.euler.x - 
		   SDI_F->current_position.mainToOrigin.euler.x) +
	       SQR(current_position.mainToOrigin.euler.y - 
		   SDI_F->current_position.mainToOrigin.euler.y)),
	  RAD_TO_DEG(current_position.mainToOrigin.euler.yaw - 
		     SDI_F->current_position.mainToOrigin.euler.yaw));
#endif
  return OK;
}

static void updateCurrentTrajectory(int trajectoryLength) {
  int i;
  if (currentTraj) free(currentTraj);
  currentTraj = (GENPOS_CART_CONFIG*) malloc (trajectoryLength *
					      sizeof(GENPOS_CART_CONFIG));
  currentTraj[0].x = nddPiloTrajectory.header.x0;
  currentTraj[0].y = nddPiloTrajectory.header.y0;
  currentTraj[0].theta = nddPiloTrajectory.header.theta0;
  for (i=1; i<trajectoryLength; i++) {
    currentTraj[i].theta = currentTraj[i-1].theta 
      + nddPiloTrajectory.seg[i-1].dTheta;
    currentTraj[i].x = currentTraj[i-1].x 
      + nddPiloTrajectory.seg[i-1].dL * cos(currentTraj[i].theta);
    currentTraj[i].y = currentTraj[i-1].y
      + nddPiloTrajectory.seg[i-1].dL * sin (currentTraj[i].theta);
  }
}

static int updateTrajGoalPos() {

  static int oldNumTraj = -1;
  static int nextToVisit = 0;
  static int trajectoryLength = -1;
  double xSubGoal, ySubGoal, th, distToPoint, dist;
  double dx, dy, distNext, distNextNew;

  /* read poster */
  if (!trackPosterId) {
    if (posterFind(SDI_F->trajPosterName.name, &trackPosterId) == ERROR)
      return 1;
  }
  if (nddPILO_TRAJECTORYPosterRead(trackPosterId, &nddPiloTrajectory) 
      == ERROR) {
    trackPosterId = NULL;
    fprintf(stderr, "ndd: unable to read trajectory poster\n");
    return 1;
  }
  
  /* new trajectory in poster or no more trajectory */
  if ( (oldNumTraj != nddPiloTrajectory.header.numTraj) ||
       (currentTraj == NULL) ) {
    trajectoryLength = nddPiloTrajectory.header.nbSegs+1;  
    updateCurrentTrajectory(trajectoryLength);
    oldNumTraj = nddPiloTrajectory.header.numTraj;
    nextToVisit = 1;
/*     SDI_F->global_goal.x = currentTraj[nextToVisit].x; */
/*     SDI_F->global_goal.y = currentTraj[nextToVisit].y; */
/*     SDI_F->global_goal.theta = currentTraj[nextToVisit].theta; */
  }
  
  /* select next point that is at least at distSwitchTraj from current pos */
  while ((distToPoint = getDistToPoint(currentTraj[nextToVisit].x,
				       currentTraj[nextToVisit].y))
	 <= SDI_F->params.distSwitchTraj &&
	 nextToVisit < trajectoryLength) {
    nextToVisit++;
    printf ("new goal resestTestBlocked\n");
    resestTestBlocked();
  }

  if (nextToVisit >= trajectoryLength) 
    nextToVisit = trajectoryLength - 1;


  /* test if goal not far */
  distNext = getDistToPoint(currentTraj[nextToVisit].x,
			    currentTraj[nextToVisit].y);
  dist = 0;
  if (nextToVisit != 0 && distNext > NDD_DIST_MAX_SUBGOAL) {

    dx = currentTraj[nextToVisit].x-currentTraj[nextToVisit-1].x;
    dy = currentTraj[nextToVisit].y-currentTraj[nextToVisit-1].y;
    th = atan2(dy, dx);

    /* computes a nearest subgoal on the segment [next-1,next] */
    dist = distNext - NDD_DIST_MAX_SUBGOAL;
    if (dist < 0) dist = 0;
    if (dist > hypot(dy, dx)) dist = 0;

/*     min(NDD_DIST_MAX_SUBGOAL, hypot(dy, dx)); */


    xSubGoal = currentTraj[nextToVisit].x - cos(th) * dist;
    ySubGoal = currentTraj[nextToVisit].y - sin(th) * dist;
    distNextNew = getDistToPoint(xSubGoal, ySubGoal);  
/* #define CUT_TRAJ */
#ifdef CUT_TRAJ
    printf ("! GOAL TOO FAR = %5.2f %5.2f (dist = %5.2f -> new =%5.2f)  -> \n",
	    currentTraj[nextToVisit].x,
	    currentTraj[nextToVisit].y, distNext, distNextNew);
#else
    printf ("! GOAL TOO FAR. SUGGEST : %5.2f %5.2f (dist = %5.2f)\n",
	    xSubGoal,
	    ySubGoal, dist);
#endif
  }
   
#ifdef CUT_TRAJ
  else 
#endif
{
    xSubGoal = currentTraj[nextToVisit].x;
    ySubGoal = currentTraj[nextToVisit].y;
  }
    

  printf ("------ POS = %5.2f %5.2f %5.2f   GOAL = %5.2f %5.2f %5.2f\n",
	  SDI_F->current_position.mainToOrigin.euler.x,
	  SDI_F->current_position.mainToOrigin.euler.y,
	  SDI_F->current_position.mainToOrigin.euler.yaw,
	  xSubGoal, ySubGoal, currentTraj[nextToVisit].theta);

  SDI_F->global_goal.x = xSubGoal;
  SDI_F->global_goal.y = ySubGoal;
  SDI_F->global_goal.theta = currentTraj[nextToVisit].theta;    

  /* added by guillo for akin */
  SDI_F->nextToVisit = nextToVisit;

  if (nextToVisit == (trajectoryLength-1) && dist == 0)
      return 1;

  return 0;
}

/* ----------------------------------------------------------------------
 *
 * getLocalGoal
 *
 */

static BOOL getLocalGoal(NDD_POINT * pt) 
{
  POM_EULER_V m2o;
  double x0, y0, theta0, st, ct;
  
  m2o = SDI_F->current_position.mainToOrigin;
  
  pomEulerVInvert(&m2o);

  x0 = m2o.euler.x;
  y0 = m2o.euler.y;
  theta0 = m2o.euler.yaw;
  ct = cos(theta0);
  st = sin(theta0);
  
  pt->x = x0 + SDI_F->global_goal.x * ct - SDI_F->global_goal.y * st;
  pt->y = y0 + SDI_F->global_goal.y * ct + SDI_F->global_goal.x * st;

  fprintf(stdout, "ndd: LOCAL goal  x: %5.2f  y: %5.2f\n", pt->x, pt->y);
    
  SDI_F->local_goal.x = pt->x;
  SDI_F->local_goal.y = pt->y;
  SDI_F->local_goal.theta = 0;

  return TRUE;
}


static double getDistToPoint(double x, double y) {
  double dx,dy;

  dx = x - SDI_F->current_position.mainToOrigin.euler.x;
  dy = y - SDI_F->current_position.mainToOrigin.euler.y;
  
  return (sqrt (dx*dx + dy*dy));
}

static double getDistToGoal()
{
  double dx,dy;
  double d;

  dx = SDI_F->global_goal.x - 
       SDI_F->current_position.mainToOrigin.euler.x;
  dy = SDI_F->global_goal.y - 
       SDI_F->current_position.mainToOrigin.euler.y;
  
  d = sqrt (dx*dx + dy*dy);
 
  fprintf(stdout, "ndd: dist to goal : %5.2f\n", d);

  return d;
}


/* ----------------------------------------------------------------------
 *
 * setRobotOrientation
 *
 */
static void setRobotOrientation()
{
  double diff;
  double ori;

  ori = SDI_F->current_position.mainToOrigin.euler.yaw;
  ori = angleLimit(ori);

  diff = angleLimit(SDI_F->global_goal.theta - ori);
  
  if ( diff>=0)
    SDI_F->speedRef.w = SDI_F->maxSpeed.angular*0.8;
  else 
    SDI_F->speedRef.w = - SDI_F->maxSpeed.angular*0.8;

  SDI_F->speedRef.numRef++;
  SDI_F->speedRef.v = 0.0;
  SDI_F->speedRef.vt = 0.0;

  printf ("setRobotOrientation: w=%g\n",SDI_F->speedRef.w);
  nddDiagramnddDiagramPosterPosterWrite(NDD_DIAGRAM_POSTER_ID, &diagrams);
}
  
/* ----------------------------------------------------------------------
 *
 * updateObstacles - read obstacles
 *
 */

static STATUS updateObstacles(int *bilan) {
  if (!obstPosterId) {
    if (posterFind(SDI_F->inputPosters.obstPosterName, 
		   &obstPosterId) == ERROR) {
      fprintf(stdout, "ndd: posterFind(%s) failed\n",
	      SDI_F->inputPosters.obstPosterName);
      *bilan = S_ndd_E_CANNOT_READ_OBSTACLES;
      return ERROR;
    }
  }
  if (readObstPoster(obstPosterId, 
		       SDI_F->inputPosters.obstType, bilan) != OK) {
    nddStopRequired = TRUE;
    nddState = NDD_FINISHING;
    return ERROR;
  }
  return OK;
}

/* ----------------------------------------------------------------------
 *
 * nddGotoLocal
 *
 */

static STATUS nddGotoLocal(int * bilan, double vmax, double wmax) 
{
  NDD_POINT * sourcePoints;
  NDD_SPEED_REF local_speedRef;
  NDD_POINT local_goal;
  struct timeval tv1,tv2;
  double distToGoal;
  int numEnlargement;
  static int prevNumEnlarg=-1;
  double v;
  int i;

  if ( (SDI_F->inputPosters.obstType == NDD_INPUT_CART_POINTS) ||
       (SDI_F->inputPosters.obstType == NDD_INPUT_POLAR_POINTS) ||
       (SDI_F->inputPosters.obstType == NDD_INPUT_DEBUG) )
    sourcePoints = (NDD_POINT*) nddSickPoster.Points;

  getLocalGoal(&local_goal);
  distToGoal = sqrt(local_goal.x*local_goal.x+local_goal.y*local_goal.y);

  gettimeofday(&tv1,NULL);
  
  v = SDI_F->speedRef.v;
  numEnlargement = (int) ((v / SDI_F->maxSpeed.linear * (NB_ENLARGMENTS - 1)) 
			  + 0.5);
  if (numEnlargement != prevNumEnlarg) {
    printf("Enlargement : %d -> %d\n",
	   numEnlargement, prevNumEnlarg);
    prevNumEnlarg = numEnlargement;
  }
  if (numEnlargement < 0) {
    printf ("** ndd BUG numEnlargement = %d\n", numEnlargement);
    numEnlargement = 0;
  }
  if (numEnlargement >= NB_ENLARGMENTS) {
    printf ("** ndd BUG numEnlargement = %d\n", numEnlargement);
    numEnlargement = NB_ENLARGMENTS -1;
  }

  switch(SDI_F->inputPosters.obstType) {
  case NDD_INPUT_POLAR_POINTS:
  case NDD_INPUT_CART_POINTS:
  case NDD_INPUT_DEBUG:
    if (!nddComputeRefFromPoints( &diagrams, sourcePoints, 
				  nddSickPoster.Header.np,
				  SDI_F->params.dMaxLocalMap,
				  enlargement[numEnlargement],
				  SDI_F->params.securityDistance, 
				  &local_goal, 
				  SDI_F->params.smax, 
				  vmax,
				  wmax,
				  &SDI_F->motionChoice.motionDirection,
				  &local_speedRef, SDI_F->params.linFactor)) {
      printf ("nddGotoLocal: pb in nddComputeRefFromPoints\n");
    }
    break;

  case NDD_INPUT_CART_SEGS:
  case NDD_INPUT_POLAR_SEGS:
    if (!nddComputeRefFromSegments( &diagrams,
				    (NDD_SEGMENT*) (&nddPolarAspectPoster.PolarSegs), 
				    nddPolarAspectPoster.Header.nbSegs,
				    (NDD_POINT*) (&nddPolarAspectPoster.Points),
				    nddPolarAspectPoster.Header.nbLonelyPoints,
				    SDI_F->params.dMaxLocalMap,
				    enlargement[numEnlargement],
				    SDI_F->params.securityDistance, 
				    &local_goal, 
				    SDI_F->params.smax, 
				    vmax,
				    wmax,
				    &SDI_F->motionChoice.motionDirection,
				    &local_speedRef, SDI_F->params.linFactor)){
      printf ("nddGotoLocal: pb in nddComputeRefFromSegments\n");
    }
    break;

  default:
    *bilan = S_ndd_E_CANNOT_READ_OBSTACLES;
    return ERROR;
  }

  gettimeofday(&tv2,NULL);
/*   fprintf(stdout, "ndd: computeRef took %ld secs and %ld msecs\n",  */
/* 	  tv2.tv_sec-tv1.tv_sec, tv2.tv_usec-tv1.tv_usec); */
  
  /*send new ref to SDI*/
/*   fprintf(stdout, "ndd: v=%f   w=%f\n", */
/* 	  local_speedRef.v, local_speedRef.w); */
  
/*   fprintf(stdout, "ndd: numRef : %d\n", SDI_F->speedRef.numRef); */
/*   fprintf(stdout, "ndd: vmax : %f\n", SDI_F->speedRef.vmax); */
/*   fprintf(stdout, "ndd: wmax : %f\n", SDI_F->speedRef.wmax); */
/*   fprintf(stdout, "ndd: linAccMax : %f\n", SDI_F->speedRef.linAccelMax); */
/*   fprintf(stdout, "ndd: angAccMax : %f\n", SDI_F->speedRef.angAccelMax); */
  
  filterSpeed(&local_speedRef, &SDI_F->speedRef);

  SDI_F->speedRef.numRef++;
  SDI_F->speedRef.w = local_speedRef.w;
  SDI_F->speedRef.v = local_speedRef.v;
  SDI_F->speedRef.vt = 0.0;

  SDI_F->motionChoice.valley = nddGiveAngleFromSector(diagrams.selected_valley, 
						      diagrams.nsectors);
  if (SDI_F->maxSpeed.linear)
    SDI_F->motionChoice.speedPercent = SDI_F->speedRef.v/SDI_F->maxSpeed.linear;
  else
    SDI_F->motionChoice.speedPercent = SDI_F->speedRef.v;

  nddDiagramnddDiagramPosterPosterWrite(NDD_DIAGRAM_POSTER_ID, &diagrams);
  return OK;
}


static void filterSpeed(NDD_SPEED_REF *wantedSpeed, GENPOS_CART_SPEED *oldSpeed)
{
  double vdiff, wdiff;
  double period;

  period = EXEC_TASK_PERIOD(NDD_EXECTASK_NUM);

  vdiff = wantedSpeed->v - oldSpeed->v;
  wdiff = wantedSpeed->w - oldSpeed->w;

  if (fabs(vdiff) / period > oldSpeed->linAccelMax)
    wantedSpeed->v = oldSpeed->v + SIGN(vdiff) * oldSpeed->linAccelMax * period;
  if (fabs(wantedSpeed->v) > oldSpeed->vmax)
    wantedSpeed->v = SIGN(vdiff) * oldSpeed->vmax;

  if (fabs(wdiff) / period > oldSpeed->angAccelMax)
    wantedSpeed->w = oldSpeed->w + SIGN(wdiff) * oldSpeed->angAccelMax * period;
  if (fabs(wantedSpeed->w) > oldSpeed->wmax)
    wantedSpeed->w = SIGN(wdiff) * oldSpeed->wmax;
}

/*------------------------------------------------------------------------
 * nddReport
 *
 * Description: 
 *
 */

static void nddReport(int *bilan) 
{
  if (nddWasBlocked) {
    printf ("NDD: WAS BLOCKED\n");
    *bilan = S_ndd_E_BLOCKED;
  }
  if (nddStopRequired) {
    printf ("NDD: WAS STOPPED\n");
    *bilan = nddStopReport;
  }
  printf ("report resestTestBlocked\n");
  resestTestBlocked();
  nddWasBlocked = FALSE;
}


/*------------------------------------------------------------------------
 * resestTestBlocked
 *
 * Description: 
 *
 */

static void   resestTestBlocked(void) {
  testBlockedDuration = 0;
}

/*------------------------------------------------------------------------
 * testBlocked
 *
 * Description: 
 *
 */
static BOOL testBlocked(void)
{
  static double prevDistToGoal;
  double currentDistToGoal;
  double moveToGoal;
  BOOL blocked=FALSE;

  /* reinit test */
  if (testBlockedDuration == 0) {
    prevDistToGoal = getDistToGoal();
    testBlockedDuration++;
    printf ("NDD: test block init distToGoal(%5.2f %5.2f) = %5.2f\n", 
	    SDI_F->global_goal.x, SDI_F->global_goal.y,
	    prevDistToGoal);
  }

  /* time to test (nb iterations) */
  else if (testBlockedDuration >= 
	   (int) ((double)NDD_BLOCKED_THRESHOLD_NSECS / 
		  EXEC_TASK_PERIOD(NDD_EXECTASK_NUM))) {
    currentDistToGoal = getDistToGoal();
    moveToGoal = prevDistToGoal-currentDistToGoal;
    
    if (moveToGoal < NDD_BLOCKED_THRESHOLD_METERS) {
      nddState = NDD_BLOCKED;
      nddWasBlocked = TRUE;
      blocked = TRUE;
    }
    else {
      resestTestBlocked();
    }
    printf ("NDD: test block after %d iter : moveToGoal(%5.2f %5.2f) %5.2f < %5.2f ? : %s\n", 
	    testBlockedDuration, 
	    SDI_F->global_goal.x, SDI_F->global_goal.y,
	    moveToGoal, 
	    NDD_BLOCKED_THRESHOLD_METERS,
	    blocked ? " BLOCKED !!!\n":"ok");
  }

  /* increment time */
  else {
    printf ("NDD: test block %d iter\n",
	    testBlockedDuration);
    testBlockedDuration++;

    /* if robot is moving, reinit test to be ready when robot is blocked */
    currentDistToGoal = getDistToGoal();
    moveToGoal = prevDistToGoal-currentDistToGoal;
    if (moveToGoal > NDD_BLOCKED_THRESHOLD_METERS) 
      resestTestBlocked();
  }

  return blocked;
}

  

